import re
import logging

from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event

from liiatools.datasets.shared_functions import common

log = logging.getLogger(__name__)


def add_year_column(stream, input, la_log_dir):
    """
    Searches the filename for the year by finding any four-digit number starting with 20

    :param stream: A filtered list of event objects of type StartTable
    :return: An updated list of event objects
    """
    year = None
    for event in stream:
        if isinstance(event, events.StartTable):
            try:
                file_dir = event.filename
                year = common.check_year(file_dir)
                yield event.from_event(event, year=year)
            except AttributeError:
                common.save_year_error(input, la_log_dir)
        elif isinstance(event, events.EndTable):
            yield event
            year = None
        elif isinstance(event, events.EndRow) and year is not None:
            yield event.from_event(event, year=year)
        else:
            yield event


@streamfilter(check=lambda x: x.get("header") in ["CHILD"], fail_function=pass_event)
def create_la_child_id(event, la_code):
    """
    Creates an identifier from a combination of the Child Unique ID and Local Authority so matching child IDs
    are not removed in the merging a de-duping steps

    :param event: A filtered list of event objects
    :param la_code: The 3-character LA code used to identify a local authority
    :return: An updated list of event objects
    """
    if isinstance(event.cell, str) and event.cell[-2:] == ".0":
        child_id = int(float(event.cell))
    elif isinstance(event.cell, float):
        child_id = int(event.cell)
    else:
        child_id = event.cell

    la_child_id = f"{child_id}_{la_code}"
    yield event.from_event(event, cell=la_child_id)



import logging

from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from sfdata_stream_parser.checks import type_check

from liiatools.datasets.annex_a.lds_annexa_clean.converters import to_integer
from liiatools.datasets.shared_functions.common import check_postcode
from liiatools.datasets.shared_functions.converters import to_date
from liiatools.datasets.annex_a.lds_annexa_clean.regex import parse_regex

log = logging.getLogger(__name__)


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def clean_cell_category(event):
    """
    Checks the values of a cell against the config file using code, name and regex matching
    returns the matched value for a successful match and blank if not
    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """

    try:
        for c in event.category_config:
            if event.value:
                if str(c["code"]).lower() in str(event.value).lower():
                    return event.from_event(
                        event, value=c["code"], formatting_error="0"
                    )
                elif str(c["name"]).lower() == str(event.value).lower():
                    return event.from_event(
                        event, value=c["code"], formatting_error="0"
                    )

                for r in c.get("regex", []):
                    p = parse_regex(r)
                    if p.match(str(event.value)) is not None:
                        return event.from_event(
                            event, value=c["code"], formatting_error="0"
                        )

            else:
                return event.from_event(event, value="", formatting_error="0")

        else:
            return event.from_event(event, value="", formatting_error="1")
    except (
        AttributeError,
        KeyError,
    ):  # Raised in case there is no config item for the given cell
        return event


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def clean_integers(event):
    """
    Convert all values that should be integers into integers based on the annex-a-merge.yaml file
    if they cannot be converted record this as event.error = 1
    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    try:
        numeric = event.other_config["type"]
    except (
        AttributeError,
        KeyError,
    ):  # Raised in case there is no config item for the given cell
        return event

    if numeric == "integer":
        try:
            text = to_integer(event.value)
            return event.from_event(event, value=text, formatting_error="0")
        except (AttributeError, TypeError, ValueError):
            return event.from_event(event, value="", formatting_error="1")
    else:
        return event


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def clean_dates(event):
    """
    Convert all values that should be dates to dates based on the annex-a-merge.yaml file
    if they cannot be converted record this as event.error = 1
    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    try:
        date = event.other_config["type"]
    except (
        AttributeError,
        KeyError,
    ):  # Raised in case there is no config item for the given cell
        return event
    if date == "date":
        try:
            text = to_date(event.value)
            return event.from_event(event, value=text, formatting_error="0")
        except (AttributeError, TypeError, ValueError):
            return event.from_event(event, value="", formatting_error="1")
    else:
        return event


@streamfilter(
    check=lambda x: x.get("column_header") in ["Placement postcode"],
    fail_function=pass_event,
)
def clean_postcodes(event):
    """
    Check that all values that should be postcodes are postcodes
    if they are not postcodes record this as event.error = 1
    :param event: A filtered list of event objects which have a column header of "Placement postcode"
    :return: An updated list of event objects
    """
    error = "0"
    text = ""
    try:
        text = check_postcode(event.value)
    except (AttributeError, TypeError, ValueError):
        error = "1"
    return event.from_event(event, value=text, formatting_error=error)


def clean(stream):
    """
    Compile cleaning functions
    :param stream: A filtered list of event objects
    :return: An updated list of event objects
    """
    stream = clean_cell_category(stream)
    stream = clean_integers(stream)
    stream = clean_dates(stream)
    stream = clean_postcodes(stream)
    return stream



from pathlib import Path
import yaml
import logging
import click_log

# dependencies for cleanfile()
from liiatools.datasets.s903.lds_ssda903_clean import (
    configuration as clean_config,
    parse,
    populate,
    filters,
    degrade,
    logger,
    file_creator,
    prep,
)

# dependencies for la_agg()
from liiatools.datasets.s903.lds_ssda903_la_agg import configuration as agg_config
from liiatools.datasets.s903.lds_ssda903_la_agg import process as agg_process

# dependencies for pan_agg()
from liiatools.datasets.s903.lds_ssda903_pan_agg import configuration as pan_config
from liiatools.datasets.s903.lds_ssda903_pan_agg import process as pan_process

# dependencies for sufficiency_output()
from liiatools.datasets.s903.lds_ssda903_sufficiency import configuration as suff_config
from liiatools.datasets.s903.lds_ssda903_sufficiency import process as suff_process

from liiatools.spec import common as common_asset_dir
from liiatools.datasets.shared_functions.common import (
    flip_dict,
    check_file_type,
    supported_file_types,
)

log = logging.getLogger()
click_log.basic_config(log)

COMMON_CONFIG_DIR = Path(common_asset_dir.__file__).parent
# Get all the possible LA codes that could be used
with open(f"{COMMON_CONFIG_DIR}/LA-codes.yml") as las:
    la_list = list(yaml.full_load(las)["data_codes"].values())


def cleanfile(input, la_code, la_log_dir, output):
    """
    Cleans input SSDA903 csv files according to config and outputs cleaned csv files.
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param la_log_dir: should specify the path to the local authority's log folder
    :param output: should specify the path to the output folder
    :return: None
    """

    # Prepare file
    # List of commonly submitted unneeded files
    drop_file_list = [
        "Extended Review",
        "Pupil Premium Children",
        "Children Ceasing to be looked after for other reasons",
        "Distance and Placement Extended",
        "Extended Adoption",
        "Children Ceased Care During the Year",
        "Children Looked After on 31st March",
        "Children Started Care During the Year",
    ]
    prep.delete_unrequired_files(
        input, drop_file_list=drop_file_list, la_log_dir=la_log_dir
    )
    if prep.check_blank_file(input, la_log_dir=la_log_dir) == "empty":
        return
    prep.drop_empty_rows(input, input)

    # Configuration
    config = clean_config.Config()
    la_name = flip_dict(config["data_codes"])[la_code]
    if (
        check_file_type(
            input,
            file_types=[".csv"],
            supported_file_types=supported_file_types,
            la_log_dir=la_log_dir,
        )
        == "incorrect file type"
    ):
        return

    # Open & Parse file
    stream = parse.parse_csv(input=input)
    stream = populate.add_year_column(stream, input=input, la_log_dir=la_log_dir)

    # Configure stream
    stream = clean_config.configure_stream(stream, config)

    # Clean stream
    stream = filters.clean(stream)
    stream = degrade.degrade(stream)
    stream = logger.log_errors(stream)
    stream = populate.create_la_child_id(stream, la_code=la_code)

    # Output result
    stream = file_creator.save_stream(stream, la_name=la_name, output=output)
    stream = logger.save_errors_la(stream, la_log_dir=la_log_dir)
    list(stream)


def la_agg(input, output):
    """
    Joins data from newly cleaned SSDA903 file (output of cleanfile()) to existing SSDA903 data for the depositing local authority
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param output: should specify the path to the output folder
    :return: None
    """

    # Configuration
    config = agg_config.Config()

    # Open file as DataFrame and match file type
    s903_df = agg_process.read_file(input)
    column_names = config["column_names"]
    table_name = agg_process.match_load_file(s903_df, column_names)

    # Merge file with existing file of the same type in LA output folder
    s903_df = agg_process.merge_la_files(output, s903_df, table_name)

    # De-duplicate and remove old data according to schema
    dates = config["dates"]
    s903_df = agg_process.convert_datetimes(s903_df, dates, table_name)
    sort_order = config["sort_order"]
    dedup = config["dedup"]
    s903_df = agg_process.deduplicate(s903_df, table_name, sort_order, dedup)
    s903_df = agg_process.remove_old_data(s903_df, years=6)

    # If file still has data, after removing old data: re-format and export merged file
    if len(s903_df) > 0:
        s903_df = agg_process.convert_dates(s903_df, dates, table_name)
        agg_process.export_la_file(output, table_name, s903_df)


def pan_agg(input, la_code, output):
    """
    Joins data from newly merged SSDA903 file (output of la-agg()) to existing pan-London SSDA903 data
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param output: should specify the path to the output folder
    :return: None
    """

    # Configuration
    config = pan_config.Config()

    # Read file and match type
    s903_df = pan_process.read_file(input)
    column_names = config["column_names"]
    table_name = pan_process.match_load_file(s903_df, column_names)

    # Remove unwanted datasets and merge wanted with existing output
    pan_data_kept = config["pan_data_kept"]
    if table_name in pan_data_kept:
        la_name = flip_dict(config["data_codes"])[la_code]
        s903_df = pan_process.merge_agg_files(output, table_name, s903_df, la_name)
        pan_process.export_pan_file(output, table_name, s903_df)


def sufficiency_output(input, output):
    """
    Applies data minimisation to data from aggregated SSDA903 file (output of pan-agg()) and outputs to Sufficiency analysis folder
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param output: should specify the path to the output folder
    :return: None
    """

    # Configuration
    config = suff_config.Config()

    # Read file and match type
    s903_df = suff_process.read_file(input)
    column_names = config["column_names"]
    table_name = suff_process.match_load_file(s903_df, column_names)

    # Minimises output following schema and exports file
    suff_data_kept = config["suff_data_kept"]
    if table_name in suff_data_kept:
        minimise = config["minimise"]
        s903_df = suff_process.data_min(s903_df, minimise, table_name)
        suff_process.export_suff_file(output, table_name, s903_df)



import tablib
import functools
import logging
import os

from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from sfdata_stream_parser.checks import type_check

log = logging.getLogger(__name__)


class RowEvent(events.ParseEvent):
    pass


def save_stream(stream, la_name, output):
    """
    Outputs stream to file
    :param stream: The stream to output
    :param la_name: Full name of the LA
    :param output: Location to write the output
    :return: Updated stream
    """
    stream = coalesce_row(stream)
    stream = filter_rows(stream)
    stream = create_tables(stream, la_name=la_name)
    stream = save_tables(stream, output=output)
    return stream


def coalesce_row(stream):
    """
    Create a dictionary of the column headers and cell values for a whole row, excluding those with unknown
    column headers
    For duplicate column headers, add a number to the end of the column header, so it can still be saved and
    output in the final file
    :param stream: The stream to output
    :return: Updated stream
    """
    row = None
    duplicate_column = []
    for event in stream:
        if isinstance(event, events.StartRow):
            row = {}
        elif isinstance(event, events.EndRow):
            yield RowEvent.from_event(event, row=row)
            row = None
        elif (
            row is not None
            and isinstance(event, events.Cell)
            and event.column_header != "Unknown"
        ):
            if event.column_header in row:
                duplicate_column.append(event.column_header)
                row[f"{event.column_header} {+len(duplicate_column)+1}"] = event.value
            else:
                row[event.column_header] = event.value
        else:
            yield event


@streamfilter(
    check=type_check(RowEvent), fail_function=pass_event, error_function=pass_event
)
def filter_rows(event):
    """
    Filter out all the rows that contain blank values in columns that need to be populated for data retention
    responsibilities
    :param event: The stream to output
    :return: Updated stream
    """
    # May have to move this somewhere else once I better understand the pattern
    event_types = {
        "List 1": "Date of Contact",
        "List 3": "Date of referral",
        "List 5": "Strategy discussion initiating Section 47 Enquiry Start Date",
        "List 9": "Date of Birth",
        "List 10": "Date of Decision that Child Should be Placed for Adoption",
        "List 11": "Date enquiry received",
    }

    if event.sheet_name in event_types:
        if event.row[event_types[event.sheet_name]]:
            yield event.from_event(event, filter=0)
        else:
            yield event.from_event(event, filter=1)
    else:
        yield event.from_event(event, filter=0)


class TableEvent(events.ParseEvent):
    pass


def create_tables(stream, la_name):
    """
    Append all the rows for a given table to create one concatenated data event for tables with column headers
    that matched the config file
    Ignore any tables that did not have column headers matching the config file
    :param stream: The stream to output
    :param la_name: The name of the local authority
    :return: Updated stream
    """
    data = None
    for event in stream:
        if isinstance(event, events.StartTable):
            try:
                new_headers = event.matched_column_headers + ["LA"]
                data = tablib.Dataset(headers=new_headers)
            except AttributeError:
                data = None
        elif isinstance(event, events.EndTable):
            yield event
            yield TableEvent.from_event(event, data=data)
            data = None
        elif data is not None and isinstance(event, RowEvent):
            if event.filter == 0:
                data.append(list(event.row.values()) + [la_name])
            elif event.filter == 1:
                yield event
        yield event


def save_tables(stream, output):
    """
    Save the data events as Excel files in the output directory as long as there are exactly 11 sheets
    :param stream: The stream to output
    :param output: The location of the output file
    :return: updated stream object.
    """
    book = tablib.Databook()
    sheet_name = ""
    for event in stream:
        if isinstance(event, events.StartContainer):
            book = tablib.Databook()
        elif isinstance(event, events.EndContainer) and len(book.sheets()) == 11:
            with open(f"{os.path.join(output, event.filename)}_clean.xlsx", "wb") as f:
                f.write(book.export("xlsx"))
        elif isinstance(event, events.StartTable):
            try:
                sheet_name = event.sheet_name
            except AttributeError:
                sheet_name = ""
        elif isinstance(event, TableEvent) and event.data is not None:
            dataset = event.data
            dataset.title = sheet_name
            book.add_sheet(dataset)
        yield event


@functools.cache
def lookup_column_config(table_name, column_name):
    return None



import logging
import click_log
import click as click
from pathlib import Path
import yaml


# Dependencies for cleanfile()
from sfdata_stream_parser.stream import events
from liiatools.csdatatools.util.xml import dom_parse
from liiatools.datasets.cin_census.lds_cin_clean.schema import Schema
from liiatools.csdatatools.datasets.cincensus import filters

from liiatools.datasets.cin_census.lds_cin_clean import (
    file_creator,
    configuration as clean_config,
    logger,
    validator,
    cin_record,
    converter,
)
from liiatools.spec import common as common_asset_dir
from liiatools.datasets.shared_functions.common import (
    flip_dict,
    check_file_type,
    supported_file_types,
)

# Dependencies for la_agg()
from liiatools.datasets.cin_census.lds_cin_la_agg import configuration as agg_config
from liiatools.datasets.cin_census.lds_cin_la_agg import process as agg_process

# Dependencies for pan_agg()
from liiatools.datasets.cin_census.lds_cin_pan_agg import configuration as pan_config
from liiatools.datasets.cin_census.lds_cin_pan_agg import process as pan_process

log = logging.getLogger()
click_log.basic_config(log)

COMMON_CONFIG_DIR = Path(common_asset_dir.__file__).parent
# Get all the possible LA codes that could be used
with open(f"{COMMON_CONFIG_DIR}/LA-codes.yml") as las:
    la_list = list(yaml.full_load(las)["data_codes"].values())


@click.group()
def cin_census():
    """Functions for cleaning, minimising and aggregating CIN Census files"""
    pass


@cin_census.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--la_code",
    required=True,
    type=click.Choice(la_list, case_sensitive=False),
    help="A three letter code, specifying the local authority that deposited the file",
)
@click.option(
    "--la_log_dir",
    required=True,
    type=str,
    help="A string specifying the location that the log files for the LA should be output, usable by a pathlib Path function.",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
@click_log.simple_verbosity_option(log)
def cleanfile(input, la_code, la_log_dir, output):
    """
    Cleans input CIN Census xml files according to config and outputs cleaned csv files.
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param la_log_dir: should specify the path to the local authority's log folder
    :param output: should specify the path to the output folder
    :return: None
    """

    # Open & Parse file
    if (
        check_file_type(
            input,
            file_types=[".xml"],
            supported_file_types=supported_file_types,
            la_log_dir=la_log_dir,
        )
        == "incorrect file type"
    ):
        return
    stream = dom_parse(input)
    stream = list(stream)

    # Configure stream
    config = clean_config.Config()
    la_name = flip_dict(config["data_codes"])[la_code]
    stream = filters.strip_text(stream)
    stream = filters.add_context(stream)
    stream = filters.add_schema(stream, schema=Schema().schema)
    stream = logger.inherit_LAchildID(stream)

    # Validate stream
    field_error = []
    LAchildID_error = []
    stream = validator.validate_elements(
        stream, LAchildID_error=LAchildID_error, field_error=field_error
    )
    value_error = []
    structural_error = []
    blank_error = []
    stream = logger.counter(
        stream,
        counter_check=lambda e: isinstance(e, events.StartElement)
        and hasattr(e, "valid"),
        value_error=value_error,
        structural_error=structural_error,
        blank_error=blank_error,
    )

    # Clean stream
    stream = converter.convert_true_false(stream)
    tags = [
        "LAchildID",
        "UPN",
        "FormerUPN",
        "UPNunknown",
        "PersonBirthDate",
        "GenderCurrent",
        "PersonDeathDate",
        "Ethnicity",
        "Disability",
        "CINreferralDate",
        "ReferralSource",
        "PrimaryNeedCode",
        "CINclosureDate",
        "ReasonForClosure",
        "DateOfInitialCPC",
        "AssessmentActualStartDate",
        "AssessmentInternalReviewDate",
        "AssessmentAuthorisationDate",
        "AssessmentFactors",
        "CINPlanStartDate",
        "CINPlanEndDate",
        "S47ActualStartDate",
        "InitialCPCtarget",
        "DateOfInitialCPC",
        "ICPCnotRequired",
        "ReferralNFA",
        "CPPstartDate",
        "CPPendDate",
        "InitialCategoryOfAbuse",
        "LatestCategoryOfAbuse",
        "NumberOfPreviousCPP",
        "CPPreviewDate",
    ]
    stream = validator.remove_invalid(stream, tag_list=tags)

    # Output result
    stream = cin_record.message_collector(stream)
    data = cin_record.export_table(stream)
    data = file_creator.add_fields(input, data, la_name, la_log_dir, la_code)
    file_creator.export_file(input, output, data)
    logger.save_errors_la(
        input,
        value_error=value_error,
        structural_error=structural_error,
        LAchildID_error=LAchildID_error,
        field_error=field_error,
        blank_error=blank_error,
        la_log_dir=la_log_dir,
    )


@cin_census.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--flat_output",
    required=True,
    type=str,
    help="A string specifying the directory location for the main flatfile output",
)
@click.option(
    "--analysis_output",
    required=True,
    type=str,
    help="A string specifying the directory location for the additional analysis outputs",
)
def la_agg(input, flat_output, analysis_output):
    """
    Joins data from newly cleaned CIN Census file (output of cleanfile()) to existing CIN Census data for the depositing local authority
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param flat_output: should specify the path to the folder for the main flatfile output
    :param analysis_output: should specify the path to the folder for the additional analytical outputs
    :return: None
    """

    # Configuration
    config = agg_config.Config()

    # Open file as Dataframe
    dates = config["dates"]
    flatfile = agg_process.read_file(input, dates)

    # Merge with existing data, de-duplicate and apply data retention policy
    flatfile = agg_process.merge_la_files(flat_output, dates, flatfile)
    sort_order = config["sort_order"]
    dedup = config["dedup"]
    flatfile = agg_process.deduplicate(flatfile, sort_order, dedup)
    flatfile = agg_process.remove_old_data(flatfile, years=6)

    # Output flatfile
    agg_process.export_flatfile(flat_output, flatfile)

    # Create and output factors file
    factors = agg_process.filter_flatfile(
        flatfile, filter="AssessmentAuthorisationDate"
    )
    if len(factors) > 0:
        factors = agg_process.split_factors(factors)
        agg_process.export_factfile(analysis_output, factors)

    # Create referral file
    ref, s17, s47 = agg_process.referral_inputs(flatfile)
    if len(s17) > 0 and len(s47) > 0:
        ref_assessment = config["ref_assessment"]
        ref_s17 = agg_process.merge_ref_s17(ref, s17, ref_assessment)
        ref_s47 = agg_process.merge_ref_s47(ref, s47, ref_assessment)
        ref_outs = agg_process.ref_outcomes(ref, ref_s17, ref_s47)
        agg_process.export_reffile(analysis_output, ref_outs)

    # Create journey file
    icpc_cpp_days = config["icpc_cpp_days"]
    s47_cpp_days = config["s47_cpp_days"]
    s47_j, cpp = agg_process.journey_inputs(flatfile)
    if len(s47_j) > 0 and len(cpp) > 0:
        s47_outs = agg_process.journey_merge(s47_j, cpp, icpc_cpp_days, s47_cpp_days)
        s47_day_limit = config["s47_day_limit"]
        icpc_day_limit = config["icpc_day_limit"]
        s47_journey = agg_process.s47_paths(s47_outs, s47_day_limit, icpc_day_limit)
        agg_process.export_journeyfile(analysis_output, s47_journey)


@cin_census.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--la_code",
    required=True,
    type=click.Choice(la_list, case_sensitive=False),
    help="A three letter code, specifying the local authority that deposited the file",
)
@click.option(
    "--flat_output",
    required=True,
    type=str,
    help="A string specifying the directory location for the main flatfile output",
)
@click.option(
    "--analysis_output",
    required=True,
    type=str,
    help="A string specifying the directory location for the additional analysis outputs",
)
def pan_agg(input, la_code, flat_output, analysis_output):
    """
    Joins data from newly merged CIN Census file (output of la-agg()) to existing pan-London CIN Census data and creates analytical outputs
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param flat_output: should specify the path to the folder for the main flatfile output
    :param analysis_output: should specify the path to the folder for the additional analytical outputs
    :return: None
    """

    # Configuration
    config = pan_config.Config()

    # Create flat file
    dates = config["dates"]
    flatfile = pan_process.read_file(input, dates)

    # Merge with existing pan-London data
    la_name = flip_dict(config["data_codes"])[la_code]
    flatfile = pan_process.merge_agg_files(flat_output, dates, la_name, flatfile)

    # Output flatfile
    pan_process.export_flatfile(flat_output, flatfile)

    # Create and output factors file
    factors = pan_process.filter_flatfile(
        flatfile, filter="AssessmentAuthorisationDate"
    )
    if len(factors) > 0:
        factors = pan_process.split_factors(factors)
        pan_process.export_factfile(analysis_output, factors)

    # Create referral file
    ref, s17, s47 = pan_process.referral_inputs(flatfile)
    if len(s17) > 0 and len(s47) > 0:
        ref_assessment = config["ref_assessment"]
        ref_s17 = pan_process.merge_ref_s17(ref, s17, ref_assessment)
        ref_s47 = pan_process.merge_ref_s47(ref, s47, ref_assessment)
        ref_outs = pan_process.ref_outcomes(ref, ref_s17, ref_s47)
        pan_process.export_reffile(analysis_output, ref_outs)

    # Create journey file
    icpc_cpp_days = config["icpc_cpp_days"]
    s47_cpp_days = config["s47_cpp_days"]
    s47_j, cpp = pan_process.journey_inputs(flatfile)
    if len(s47_j) > 0 and len(cpp) > 0:
        s47_outs = pan_process.journey_merge(s47_j, cpp, icpc_cpp_days, s47_cpp_days)
        s47_day_limit = config["s47_day_limit"]
        icpc_day_limit = config["icpc_day_limit"]
        s47_journey = pan_process.s47_paths(s47_outs, s47_day_limit, icpc_day_limit)
        pan_process.export_journeyfile(analysis_output, s47_journey)



from typing import Iterator
import tablib
from more_itertools import peekable

from sfdata_stream_parser import events
from sfdata_stream_parser.collectors import xml_collector


class CINEvent(events.ParseEvent):
    pass


class HeaderEvent(events.ParseEvent):
    pass


def _reduce_dict(dict_instance):
    new_dict = {}
    for key, value in dict_instance.items():
        if len(value) == 1:
            new_dict[key] = value[0]
        else:
            new_dict[key] = value
    return new_dict


@xml_collector
def text_collector(stream):
    data_dict = {}
    current_element = None
    for event in stream:
        if isinstance(event, events.StartElement):
            current_element = event.tag
        if isinstance(event, events.TextNode) and event.text:
            data_dict.setdefault(current_element, []).append(event.text)

    return _reduce_dict(data_dict)


@xml_collector
def cin_collector(stream):
    data_dict = {}
    stream = peekable(stream)
    last_tag = None
    while stream:
        event = stream.peek()
        last_tag = event.get("tag", last_tag)
        if event.get("tag") in (
            "Assessments",
            "CINPlanDates",
            "Section47",
            "ChildProtectionPlans",
        ):
            data_dict.setdefault(event.tag, []).append(text_collector(stream))
        else:
            if isinstance(event, events.TextNode) and event.text:
                data_dict.setdefault(last_tag, []).append(event.text)
            next(stream)

    return _reduce_dict(data_dict)


@xml_collector
def child_collector(stream):
    data_dict = {}
    stream = peekable(stream)
    assert stream.peek().tag == "Child"
    while stream:
        event = stream.peek()
        if event.get("tag") in ("ChildIdentifiers", "ChildCharacteristics"):
            data_dict.setdefault(event.tag, []).append(text_collector(stream))
        elif event.get("tag") == "CINdetails":
            data_dict.setdefault(event.tag, []).append(cin_collector(stream))
        else:
            next(stream)

    return _reduce_dict(data_dict)


@xml_collector
def message_collector(stream):
    stream = peekable(stream)
    assert stream.peek().tag == "Message", "Expected Message, got {}".format(
        stream.peek().tag
    )
    while stream:
        event = stream.peek()
        if event.get("tag") == "Header":
            header_record = text_collector(stream)
            if header_record:
                yield HeaderEvent(record=header_record)
        elif event.get("tag") == "Child":
            cin_record = child_collector(stream)
            if cin_record:
                yield CINEvent(record=cin_record)
        else:
            next(stream)


__EXPORT_HEADERS = [
    "LAchildID",
    "Date",
    "Type",
    "CINreferralDate",
    "ReferralSource",
    "PrimaryNeedCode",
    "CINclosureDate",
    "ReasonForClosure",
    "DateOfInitialCPC",
    "ReferralNFA",
    "CINPlanStartDate",
    "CINPlanEndDate",
    "S47ActualStartDate",
    "InitialCPCtarget",
    "ICPCnotRequired",
    "AssessmentActualStartDate",
    "AssessmentInternalReviewDate",
    "AssessmentAuthorisationDate",
    "Factors",
    "CPPstartDate",
    "CPPendDate",
    "InitialCategoryOfAbuse",
    "LatestCategoryOfAbuse",
    "NumberOfPreviousCPP",
    "UPN",
    "FormerUPN",
    "UPNunknown",
    "PersonBirthDate",
    "ExpectedPersonBirthDate",
    "GenderCurrent",
    "PersonDeathDate",
    "PersonSchoolYear",
    "Ethnicity",
    "Disabilities",
]


def _maybe_list(value):
    if value is None:
        value = []
    if not isinstance(value, list):
        value = [value]
    return value


def cin_event(record, property, event_name=None):
    if event_name is None:
        event_name = property
    value = record.get(property)
    if value:
        new_record = {**record, "Date": value, "Type": event_name}
        return ({k: new_record.get(k) for k in __EXPORT_HEADERS},)

    return ()


def event_to_records(event: CINEvent) -> Iterator[dict]:
    record = event.record
    child = {
        **record.get("ChildIdentifiers", {}),
        **record.get("ChildCharacteristics", {}),
    }
    child["Disabilities"] = ",".join(_maybe_list(child.get("Disability")))

    for cin_item in _maybe_list(record.get("CINdetails")):
        yield from cin_event({**child, **cin_item}, "CINreferralDate")
        yield from cin_event({**child, **cin_item}, "CINclosureDate")

        for assessment in _maybe_list(cin_item.get("Assessments")):
            assessment["Factors"] = ",".join(
                _maybe_list(assessment.get("AssessmentFactors"))
            )
            yield from cin_event(
                {**child, **cin_item, **assessment}, "AssessmentActualStartDate"
            )
            yield from cin_event(
                {**child, **cin_item, **assessment}, "AssessmentAuthorisationDate"
            )

        for cin in _maybe_list(cin_item.get("CINPlanDates")):
            yield from cin_event(
                {**child, **cin_item, **cin},
                "CINPlanStartDate",
            )
            yield from cin_event({**child, **cin_item, **cin}, "CINPlanEndDate")

        for s47 in _maybe_list(cin_item.get("Section47")):
            yield from cin_event({**child, **cin_item, **s47}, "S47ActualStartDate")

        for cpp in _maybe_list(cin_item.get("ChildProtectionPlans")):
            yield from cin_event({**child, **cin_item, **cpp}, "CPPstartDate")
            yield from cin_event({**child, **cin_item, **cpp}, "CPPendDate")
            for cpp_review in _maybe_list(cpp.get("CPPreviewDate")):
                cpp_review = {"CPPreviewDate": cpp_review}
                yield from cin_event(
                    {**child, **cin_item, **cpp, **cpp_review}, "CPPreviewDate"
                )


def export_table(stream):
    data = tablib.Dataset(headers=__EXPORT_HEADERS)
    for event in stream:
        if isinstance(event, CINEvent):
            for record in event_to_records(event):
                data.append([record.get(k, "") for k in __EXPORT_HEADERS])
    return data



import re
from typing import Pattern

__flag_resolved = dict(i=re.I, m=re.M, s=re.S, u=re.U, l=re.L, x=re.X)


def resolve_flags(flags: str) -> int:
    flag_expr = 0
    if flags is not None:
        for flag in flags:
            flag_expr = flag_expr | __flag_resolved[flag]
    return flag_expr


def parse_regex(regex: str) -> Pattern:
    """
    Parse a regex pattern '/{pattern}/{modifiers}'

    :param regex: regex expression
    :return: Compiled regex
    """
    separator = re.escape(regex[0])
    pattern = re.compile(
        "{separator}(.+)({separator}([imsulx]+)?)".format(separator=separator)
    )
    match = pattern.match(regex)
    if match is None:
        raise Exception("Failed to parse regular expression: '{}'".format(regex))

    pattern = match.group(1)
    flags = match.group(3)
    flags = resolve_flags(flags)

    return re.compile(pattern, flags)



import click as click

from liiatools.datasets.annex_a.annex_a_cli import annex_a
from liiatools.datasets.cin_census.cin_cli import cin_census
from liiatools.datasets.s903.s903_cli import s903


@click.group()
def cli():
    pass


cli.add_command(annex_a)
cli.add_command(cin_census)
cli.add_command(s903)

if __name__ == "__main__":
    cli()



from sfdata_stream_parser.checks import type_check
from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event


@streamfilter(check=type_check(events.TextNode), fail_function=pass_event)
def convert_true_false(event):
    """
    Search for any events that have the schema type="yesnotype" and convert any values of false to 0 and true to 1

    :param event: A filtered list of event objects
    :return: An updated list of event objects
    """
    if hasattr(event, "schema"):
        if event.schema.type.name == "yesnotype":
            if event.text.lower() == "false":
                event = event.from_event(event, text="0")
            elif event.text.lower() == "true":
                event = event.from_event(event, text="1")
    return event



import click as click
from pathlib import Path
import yaml
import logging
import click_log

# dependencies for cleanfile()
from liiatools.datasets.s903 import s903_main_functions

from liiatools.spec import common as common_asset_dir

log = logging.getLogger()
click_log.basic_config(log)

COMMON_CONFIG_DIR = Path(common_asset_dir.__file__).parent
# Get all the possible LA codes that could be used
with open(f"{COMMON_CONFIG_DIR}/LA-codes.yml") as las:
    la_list = list(yaml.full_load(las)["data_codes"].values())


@click.group()
def s903():
    """Functions for cleaning, minimising and aggregating SSDA903 files"""
    pass


@s903.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--la_code",
    required=True,
    type=click.Choice(la_list, case_sensitive=False),
    help="A three letter code, specifying the local authority that deposited the file",
)
@click.option(
    "--la_log_dir",
    required=True,
    type=str,
    help="A string specifying the location that the log files for the LA should be output, usable by a pathlib Path function.",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
@click_log.simple_verbosity_option(log)
def cleanfile(input, la_code, la_log_dir, output):
    """
    Cleans input SSDA903 csv files according to config and outputs cleaned csv files.
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param la_log_dir: should specify the path to the local authority's log folder
    :param output: should specify the path to the output folder
    :return: None
    """
    output = s903_main_functions.cleanfile(input, la_code, la_log_dir, output)
    return output


@s903.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
def la_agg(input, output):
    """
    Joins data from newly cleaned SSDA903 file (output of cleanfile()) to existing SSDA903 data for the depositing local authority
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param output: should specify the path to the output folder
    :return: None
    """
    s903_main_functions.la_agg(input, output)


@s903.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--la_code",
    required=True,
    type=click.Choice(la_list, case_sensitive=False),
    help="A three letter code, specifying the local authority that deposited the file",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
def pan_agg(input, la_code, output):
    """
    Joins data from newly merged SSDA903 file (output of la-agg()) to existing pan-London SSDA903 data
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param la_code: should be a three-letter string for the local authority depositing the file
    :param output: should specify the path to the output folder
    :return: None
    """
    s903_main_functions.pan_agg(input, la_code, output)


@s903.command()
@click.option(
    "--i",
    "input",
    required=True,
    type=str,
    help="A string specifying the input file location, including the file name and suffix, usable by a pathlib Path function",
)
@click.option(
    "--o",
    "output",
    required=True,
    type=str,
    help="A string specifying the output directory location",
)
def sufficiency_output(input, output):
    """
    Applies data minimisation to data from aggregated SSDA903 file (output of pan-agg()) and outputs to Sufficiency analysis folder
    :param input: should specify the input file location, including file name and suffix, and be usable by a Path function
    :param output: should specify the path to the output folder
    :return: None
    """
    s903_main_functions.sufficiency_output(input, output)


import datetime
import logging
import os
from pathlib import Path
import yaml
from string import Template

from liiatools.datasets.annex_a.lds_annexa_clean.regex import parse_regex
from liiatools.datasets.shared_functions.common import inherit_property
from liiatools.spec import annex_a as annex_a_asset_dir
from liiatools.spec import common as common_asset_dir

from sfdata_stream_parser import events, checks
from sfdata_stream_parser.filters.generic import streamfilter, pass_event

log = logging.getLogger(__name__)

DEFAULT_CONFIG_DIR = Path(annex_a_asset_dir.__file__).parent
COMMON_CONFIG_DIR = Path(common_asset_dir.__file__).parent


def _match_column_name(actual_value, expected_value, expected_expressions=None):
    """
    Matches an actual column name against an expected values. Can optionally take a list of expressions to test.
    Returns the expected value if a match is found, or None if no match is found.
    :param actual_value: Value that exists currently
    :param expected_value: Value that we expect
    :param expected_expressions: Optional list of (regex) expressions to test as well
    :return: The expected Value or None
    """
    assert actual_value is not None, "Must test a value"
    assert expected_value is not None, "Must test against a value"

    actual_value = actual_value.lower().strip()
    test_expected_value = expected_value.lower().strip()

    if actual_value == test_expected_value:
        return expected_value

    if expected_expressions:
        for ptn in expected_expressions:
            ptn = parse_regex(ptn)
            if ptn.match(actual_value):
                return expected_value

    return None


def configure_stream(stream, config):
    """
    Loading and matching the configuration with the loaded stream
    :param stream: Set of events to parse
    :param config: The loaded configuration
    :return: An updated set of events/stream with matched configuration
    """
    sheet_config = config["datasources"]
    cell_config = config["data_config"]
    stream = identify_blank_rows(stream)
    stream = add_sheet_name(stream, config=sheet_config)
    stream = inherit_property(stream, "sheet_name")
    stream = inherit_property(stream, "column_headers")
    stream = identify_cell_header(stream)
    stream = convert_column_header_to_match(stream, config=sheet_config)
    stream = match_property_config_to_cell(
        stream, config=cell_config, prop_name="category_config"
    )
    stream = match_property_config_to_cell(
        stream, config=sheet_config, prop_name="other_config"
    )
    return stream


def identify_blank_rows(stream):
    """
    Identify any blank rows in the dataset and assign a blank_row = "1" value to the corresponding
    StartRow, Cell and EndRow
    All Cell values between StartRow and EndRow are added to a list which is then checked to see if it contains
    only None values i.e. the row is blank. If row is blank return all events with a blank_row = "1" value,
    otherwise return all events without changes

    :param stream: A filtered list of event objects
    :return: An updated list of event objects with blank row information
    """
    row = None
    blank_row = []
    for event in stream:
        if isinstance(event, events.StartRow):
            row = [event]
        elif isinstance(event, events.EndRow):
            row.append(event)
            if blank_row == [None] * len(blank_row):
                yield from [event.from_event(event, blank_row="1") for event in row]
            else:
                yield from [event.from_event(event) for event in row]
            row = None
            blank_row = []
        elif isinstance(event, events.Cell):
            row.append(event)
            blank_row.append(event.value)
        elif row:
            row.append(event)
        else:
            yield event


@streamfilter(
    check=checks.type_check(events.StartTable),
    fail_function=pass_event,
    error_function=pass_event,
)
def add_sheet_name(event, config):
    """
    Match the loaded table against one of the Annex A sheet names using fuzzy matching with regex
    the column headers will be matched against the config, building a new list of matched headers
    which is then checked against the expected columns from the config
    If columns are matched but there are more columns than expected this will save those extra columns
    as event.extra_columns
    If no columns are matched for a table this will save the sheet name and column headers as event.match_error

    :param event: A filtered list of event objects of type StartTable
    :param config: A dictionary of keys containing sheet names and values containing headers
    :return: An updated list of event objects
    """
    for table_name, table_cfg in config.items():
        matched_names = []
        extra_columns = []

        # Creates a list of tuples holding (column_name, column_regex_list) for each configured column
        header_config = [
            (name, cfg.get("regex", [])) for name, cfg in table_cfg.items()
        ]

        for actual_column in event.column_headers:
            # Check the actual value against each of the configured columns and store values
            header_matches = [
                _match_column_name(actual_column, c[0], c[1]) for c in header_config
            ]

            # Filter checks to only those that matched
            matching_configs = [c for c in header_matches if c is not None]

            # Check if we have no, one or multiple configurations that match the actual value
            if len(matching_configs) == 0:
                extra_columns.append(actual_column)
            elif len(matching_configs) == 1:
                matched_names.append(matching_configs[0])
            else:
                raise ValueError(
                    "The actual column name matched multiple configured columns"
                )
        # If all of the expected columns are present, then we have a match
        if set(table_cfg.keys()) - set(matched_names) == set():
            return events.StartTable.from_event(
                event,
                sheet_name=table_name,
                extra_columns=extra_columns,
                matched_column_headers=matched_names,
            )
    return event


@streamfilter(check=checks.type_check(events.Cell), fail_function=pass_event)
def identify_cell_header(event):
    """
    Finds the correct column header for this Cell. The cell must have the property
    column_headers set otherwise this function returns the cell unmodified.

    Based on the cell's 'column_index' it looks up the corresponding header value in column_headers.

    Provides: column_header
    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    column_headers = event.get("column_headers")
    if column_headers:
        event = event.from_event(
            event, column_header=column_headers[event.column_index]
        )
    return event


@streamfilter(
    check=checks.type_check(events.Cell),
    fail_function=pass_event,
    error_function=pass_event,
)
def convert_column_header_to_match(event, config):
    """
    Converts the column header to the correct column header it was matched with e.g. Age -> Age of Child (Years)
    :param event: A filtered list of event objects of type Cell
    :param config: The loaded configuration to use
    :return: An updated list of event objects
    """
    column_config = config[event.sheet_name]
    for c in column_config:
        if c in str(event.column_header):
            return event.from_event(event, column_header=c)

        else:
            try:
                for r in column_config[c].get("regex", []):
                    p = parse_regex(r)
                    if p.match(str(event.column_header)) is not None:
                        return event.from_event(event, column_header=c)
            except AttributeError:  # Raised in case a config item empty which is acceptable
                pass
    return event.from_event(event, column_header="Unknown")


@streamfilter(
    check=checks.type_check(events.Cell),
    fail_function=pass_event,
    error_function=pass_event,
)
def match_property_config_to_cell(event, config, prop_name):
    """
    Match the cell to the config file given the sheet name and cell header
    the config file should be a set of dictionaries for each sheet, headers within those sheets
    and config rules for those headers

    :param event: A filtered list of event objects of type Cell
    :param config: The loaded configuration to use
    :param prop_name: The property name to match with the configuration
    :return: An updated list of event objects
    """
    try:
        sheet_config = config[event.sheet_name]
        config_dict = sheet_config[event.column_header]
        return event.from_event(event, **{prop_name: config_dict})
    except KeyError:  # Raised in case there is no property item for the given sheet name and cell header
        return event


class Config(dict):
    def __init__(self, *config_files):
        super().__init__()

        if not config_files:
            config_files = [
                "DEFAULT_DATA_SOURCES",
                "DEFAULT_DATA_MAP",
                "DEFAULT_DATA_CODES",
            ]

        for file in config_files:
            if file == "DEFAULT_DATA_SOURCES":
                file = DEFAULT_CONFIG_DIR / "annex-a-merge.yml"
            elif file == "DEFAULT_DATA_MAP":
                file = DEFAULT_CONFIG_DIR / "data-map.yml"
            elif file == "DEFAULT_DATA_CODES":
                file = COMMON_CONFIG_DIR / "LA-codes.yml"
            self.load_config(file, conditional=False)

        self["config_date"] = datetime.datetime.now().isoformat()
        try:
            self["username"] = os.getlogin()
        except OSError:
            # This happens when tests are not run under a login shell, e.g. CI pipeline
            pass

    def load_config(self, filename, conditional=False, warn=False):
        """
        Load configuration from yaml file. Any loaded configuration
        is only set if the values don't already exist in CONFIG.

        Files can contain ${} placeholders following the Python string.Template format.
        The context will include any keys already existing in the configuration, any keys
        from the current file - however, if these include placeholders, the placeholders
        will not be replaced. Finally, environment variables can be referenced with
        `os_environ_VARIABLE_NAME`.

        Keyword arguments:
        filename -- Filename to load from
        conditional -- If True, ignore file if it doesn't exist. If False, fail. (default False)
        """
        if conditional and not os.path.isfile(filename):
            if warn:
                log.warning("Missing optional file {}".format(filename))

            return

        with open(filename) as FILE:
            user_config = yaml.load(FILE, Loader=yaml.FullLoader)

        log.info(
            "Loading {} configuration values from '{}'.".format(
                len(user_config), filename
            )
        )

        environment_dict = {"os_environ_{}".format(k): v for k, v in os.environ.items()}

        variables = dict(self)
        variables.update(user_config)
        variables.update(environment_dict)

        with open(filename, "rt") as FILE:
            user_config_string = FILE.read()

        user_config_template = Template(user_config_string)
        user_config_string = user_config_template.substitute(variables)

        user_config = yaml.load(user_config_string, Loader=yaml.FullLoader)

        self.update(user_config)



from sfdata_stream_parser.events import (
    StartElement,
    EndElement,
    TextNode,
    CommentNode,
    ProcessingInstructionNode,
)

try:
    from lxml import etree
except ImportError:
    pass


def dom_parse(source, **kwargs):
    """
    Equivalent of the xml parse included in the sfdata_stream_parser package, but uses the ET DOM
    and allows direct DOM manipulation.
    """
    parser = etree.iterparse(source, events=("start", "end", "comment", "pi"), **kwargs)
    for action, elem in parser:
        if action == "start":
            yield StartElement(tag=elem.tag, attrib=elem.attrib, node=elem)
            if elem.text:
                yield TextNode(text=elem.text)
        elif action == "end":
            yield EndElement(tag=elem.tag, node=elem)
            if elem.tail:
                yield TextNode(text=elem.tail)
        elif action == "comment":
            yield CommentNode(text=elem.text, node=elem)
        elif action == "pi":
            yield ProcessingInstructionNode(name=elem.target, text=elem.text, node=elem)
        else:
            raise ValueError(f"Unknown event: {action}")


def to_xml(stream, builder: etree.TreeBuilder):
    for ev in stream:
        if isinstance(ev, StartElement):
            builder.start(ev.tag, getattr(ev, "attrs", {}))
        elif isinstance(ev, EndElement):
            builder.end(ev.tag)
        elif isinstance(ev, TextNode):
            builder.data(ev.text)
        elif isinstance(ev, CommentNode):
            builder.comment(ev.text)
        elif isinstance(ev, ProcessingInstructionNode):
            builder.pi(ev.name, ev.text)
        yield ev



from collections import Counter
from datetime import datetime
import logging
import os

from sfdata_stream_parser import events
from sfdata_stream_parser.filters.generic import streamfilter, pass_event
from sfdata_stream_parser.checks import type_check

log = logging.getLogger(__name__)


class ErrorTable(events.ParseEvent):
    pass


def create_formatting_error_count(stream):
    """
    Create a list of the column headers for cells with formatting errors (event.error = 1) for each table

    :param stream: A filtered list of event objects
    :return: An updated list of event objects with error counts
    """
    formatting_error_count = None
    table_name = None
    for event in stream:
        if isinstance(event, events.StartTable):
            formatting_error_count = []
            try:
                table_name = event.table_name
            except AttributeError:
                pass
        elif isinstance(event, events.EndTable):
            yield ErrorTable.from_event(
                event,
                table_name=table_name,
                formatting_error_count=formatting_error_count,
            )
            formatting_error_count = None
        elif (
            formatting_error_count is not None
            and table_name is not None
            and isinstance(event, events.Cell)
        ):
            try:
                if event.error == "1":
                    formatting_error_count.append(event.header)
            except AttributeError:
                pass
        yield event


@streamfilter(
    check=type_check(events.Cell), fail_function=pass_event, error_function=pass_event
)
def blank_error_check(event):
    """
    Check all the values against the config to see if they are allowed to be blank
    if they are blank but should not be, record this as event.blank_error = 1

    :param event: A filtered list of event objects of type Cell
    :return: An updated list of event objects
    """
    try:
        allowed_blank = event.config_dict["canbeblank"]
        error = getattr(event, "error", "0")
        if not allowed_blank and not event.cell and error != "1":
            return event.from_event(event, blank_error="1")
        else:
            return event
    except AttributeError:  # Raised in case there is no config item for the given cell
        pass


def create_blank_error_count(stream):
    """
    Create a list of the column headers for cells with blank fields that should not be blank (event.blank_error = 1)
    for each table

    :param stream: A filtered list of event objects
    :return: An updated list of event objects
    """
    blank_error_count = None
    for event in stream:
        if isinstance(event, events.StartTable):
            blank_error_count = []
        elif isinstance(event, events.EndTable):
            blank_error_count = None
        elif isinstance(event, ErrorTable):
            yield ErrorTable.from_event(event, blank_error_count=blank_error_count)
            blank_error_count = None
        elif blank_error_count is not None and isinstance(event, events.Cell):
            try:
                if event.blank_error == "1":
                    blank_error_count.append(event.header)
            except AttributeError:
                pass
        yield event


@streamfilter(
    check=type_check(events.StartTable),
    fail_function=pass_event,
    error_function=pass_event,
)
def create_file_match_error(event):
    """
    Add a match_error to StartTables that do not have an event.sheet_name so these errors can be written to the log.txt
    file. If there is no event.sheet_name for a given StartTable that means its headers did not match any of those
    in the config file

    :param event: A filtered list of event objects of type StartTable
    :return: An updated list of event objects
    """
    try:
        if event.table_name:
            return event
    except AttributeError:
        return event.from_event(
            event,
            match_error=f"Failed to find a set of matching columns headers for file titled "
            f"'{event.filename}' which contains column headers {event.headers} so no output has been produced",
        )
    return event


@streamfilter(
    check=type_check(events.StartTable),
    fail_function=pass_event,
    error_function=pass_event,
)
def create_extra_column_error(event):
    """
    Add a extra_column_error to StartTables that have more columns than the set of expected columns so these can be written to the log.txt

    :param event: A filtered list of event objects of type StartTable
    :return: An updated list of event objects
    """
    extra_columns = [
        item for item in event.headers if item not in event.expected_columns
    ]
    if len(extra_columns) == 0:
        return event
    else:
        return event.from_event(
            event,
            extra_column_error=f"Additional columns were found in file titled "
            f"'{event.filename}' than those expected from schema for filetype = {event.table_name}, so these columns have been removed: {extra_columns}",
        )


def save_errors_la(stream, la_log_dir):
    """
    Count the error events and save them as a text file in the Local Authority Logs directory
    only save the error events if there is at least one error in said event

    :param stream: A filtered list of event objects
    :param la_log_dir: Location to save the gathered error logs
    :return: An updated list of event objects
    """
    start_time = f"{datetime.now():%Y-%m-%dT%H%M%SZ}"
    for event in stream:
        try:
            if isinstance(event, ErrorTable) and (
                event.formatting_error_count is not None
                and event.blank_error_count is not None
                and event.table_name is not None
            ):
                if event.formatting_error_count or event.blank_error_count:
                    with open(
                        f"{os.path.join(la_log_dir, event.filename)}_error_log_{start_time}.txt",
                        "a",
                    ) as f:
                        f.write(event.table_name)
                        f.write("\n")
                        if event.formatting_error_count:
                            f.write(
                                "Number of cells that have been made blank "
                                "because they could not be formatted correctly"
                            )
                            f.write("\n")
                            counter_dict = Counter(event.formatting_error_count)
                            f.write(
                                str(counter_dict)[9:-2]
                            )  # Remove "Counter({" and "})" from string
                            f.write("\n")
                        if event.blank_error_count:
                            f.write(
                                "Number of blank cells that should have contained data"
                            )
                            f.write("\n")
                            blank_counter_dict = Counter(event.blank_error_count)
                            f.write(
                                str(blank_counter_dict)[9:-2]
                            )  # Remove "Counter({" and "})" from string
                            f.write("\n")
        except AttributeError:
            pass

        if isinstance(event, events.StartTable):
            match_error = getattr(event, "match_error", None)
            if match_error:
                with open(
                    f"{os.path.join(la_log_dir, event.filename)}_error_log_{start_time}.txt",
                    "a",
                ) as f:
                    f.write(match_error)
                    f.write("\n")
            column_error = getattr(event, "extra_column_error", None)
            if column_error:
                with open(
                    f"{os.path.join(la_log_dir, event.filename)}_error_log_{start_time}.txt",
                    "a",
                ) as f:
                    f.write(column_error)
                    f.write("\n")
        yield event


def log_errors(stream):
    """
    Compile the log error functions

    :param stream: A filtered list of event objects
    :return: An updated list of event objects
    """
    stream = blank_error_check(stream)
    stream = create_formatting_error_count(stream)
    stream = create_blank_error_count(stream)
    stream = create_file_match_error(stream)
    stream = create_extra_column_error(stream)
    return stream



import re
import logging
from pathlib import Path
from datetime import datetime

from sfdata_stream_parser import events

log = logging.getLogger(__name__)

supported_file_types = [".xml", ".csv", ".xlsx", ".xlsm"]


def flip_dict(some_dict):
    """
    Potentially a temporary function which switches keys and values in a dictionary.
    May need to rewrite LA-codes YAML file to avoid this step

    :param some_dict: A config dictionary
    :return: a reversed dictionary with keys as values and vice versa
    """
    return {value: key for key, value in some_dict.items()}


def check_postcode(postcode):
    """
    Checks that the postcodes are in the right format
    :param postcode: A string with a UK-style post code
    :return: a post code, or if incorrect a blank string
    """
    if postcode:
        match = re.search(
            r"^[A-Z]{1,2}\d[A-Z\d]? *\d[A-Z]{2}$", postcode.strip(), re.IGNORECASE
        )
        return match.group(0)
    return ""


def to_short_postcode(postcode):
    """
    Remove whitespace from the beginning and end of postcodes and the last two digits for anonymity
    return blank if not in the right format
    :param postcode: A string with a UK-style post code
    :return: a shortened post code with the area, district, and sector. The units is removed
    """
    if postcode:
        try:
            match = re.search(
                r"^[A-Z]{1,2}\d[A-Z\d]? *\d[A-Z]{2}$", postcode.strip(), re.IGNORECASE
            )
            return match.group(0)
        except AttributeError:
            return ""
    return ""


def inherit_property(stream, prop_name):
    """
    Reads a property from StartTable and sets that property (if it exists) on every event between this event
    and the next EndTable event.

    :param event: A filtered list of event objects of type StartTable
    :param prop_name: The property name to inherit
    :return: An updated list of event objects
    """
    prop_value = None
    for event in stream:
        if isinstance(event, events.StartTable):
            prop_value = getattr(event, prop_name, None)
        elif isinstance(event, events.EndTable):
            prop_value = None

        if prop_value and not hasattr(event, prop_name):
            event = event.from_event(event, **{prop_name: prop_value})

        yield event


def save_year_error(input, la_log_dir):
    """
    Save errors to a text file in the LA log directory

    :param input: The input file location, including file name and suffix, and be usable by a Path function
    :param la_log_dir: Path to the local authority's log folder
    :return: Text file containing the error information
    """
    filename = Path(input).resolve().stem
    start_time = f"{datetime.now():%d-%m-%Y %Hh-%Mm-%Ss}"
    with open(
        f"{Path(la_log_dir, filename)}_error_log_{start_time}.txt",
        "a",
    ) as f:
        f.write(
            f"Could not process '{filename}' because no year was found in the name of the file"
        )


def check_year(filename):
    """
    Check a filename to see if it contains a year, if it does, return that year
    Expected year formats within string:
        2022
        14032021
        2017-18
        201819
        2019/20
        1920
        21/22
        21-22

    :param filename: Filename that probably contains a year
    :return: Year within the string
    """
    match = re.search(r"(20)\d{2}(.*\d{2})*", filename)
    if match:
        try:
            if len(match.group(2)) == 2:
                year = match.group(1) + match.group(2)
                return year
            if len(match.group(2)) == 3:
                year = match.group(1) + match.group(2)[1:]
                return year
            if len(match.group(2)) == 4:
                year = match.group(2)
                return year
        except TypeError:
            year = match.group(0)
            return year

    fy_match = re.search(r"\d{2}(.*\d{2})", filename)
    if fy_match:
        if len(fy_match.group(1)) == 2 and int(fy_match.group(1)) == int(fy_match.group(0)[:2]) + 1:
            year = "20" + fy_match.group(1)
            return year
        if len(fy_match.group(1)) == 3 and int(fy_match.group(1)[-2:]) == int(fy_match.group(0)[:2]) + 1:
            year = "20" + fy_match.group(1)[-2:]
            return year
        else:
            raise AttributeError

    else:
        raise AttributeError


def check_file_type(input, file_types, supported_file_types, la_log_dir):
    """
    Check that the correct type of file is being used, e.g. xml. If it is then continue.
    If not then check if it is in the list of supported file types. If it is then log this error to the data processor
    If it does not match any of the expected file types then log this error to the data controller

    :param input: Location of file to be cleaned
    :param file_types: A list of the expected file type extensions e.g. [".xml", ".csv"]
    :param supported_file_types: A list of file types supported by the process e.g. [".csv", ".xlsx"]
    :param la_log_dir: Location to save the error log
    :return: Continue if correct, error log if incorrect
    """
    start_time = f"{datetime.now():%d-%m-%Y %Hh-%Mm-%Ss}"
    extension = Path(input).suffix
    filename = Path(input).resolve().stem

    disallowed_file_types = list(set(supported_file_types).difference(file_types))

    if extension in file_types:
        pass

    elif extension in disallowed_file_types:
        assert extension in file_types, f"File not in the expected {file_types} format"

    else:
        with open(
            f"{Path(la_log_dir, filename)}_error_log_{start_time}.txt",
            "a",
        ) as f:
            f.write(
                f"File: '{filename}{extension}' not in any of the expected formats (csv, xml, xlsx, xlsm)"
            )
        return "incorrect file type"



import logging
import pandas as pd
from pathlib import Path

log = logging.getLogger(__name__)


def split_file(input):
    """Reads merged LA xlsx file as dictionary of dataframes"""
    pan_dict = pd.read_excel(input, sheet_name=None, index_col=None, dtype=object)
    return pan_dict


def data_minimisation(pan_dict, minimise):
    """Reduces the sensitive information passed to the pan-London file according to schema"""
    for k in list(pan_dict):
        if k in ["List 10", "List 11"]:
            del pan_dict[k]
        else:
            df = pan_dict[k]
            df = df.drop(columns=minimise[k], axis=1)
            pan_dict[k] = df
    return pan_dict


def _merge_dfs(pan_dict, old_dict, la_name):
    """Deletes existing data for new LA from pan file
    Merges new LA data to pan file"""
    for k in pan_dict.keys():
        new_df = pan_dict[k]
        old_df = old_dict[k]
        old_df = old_df.drop(old_df[old_df["LA"] == la_name].index)
        merged_df = pd.concat([new_df, old_df], axis=0, ignore_index=True)
        pan_dict[k] = merged_df
    return pan_dict


def merge_agg_files(output, pan_dict, la_name):
    """Checks if pan file exists
    Passes old and new file to function to be merged"""
    output_file = Path(output, f"pan_London_Annex_A.xlsx")
    if output_file.is_file():
        old_dict = pd.read_excel(
            output_file, sheet_name=None, index_col=None, dtype=object
        )
        _merge_dfs(pan_dict, old_dict, la_name)
    return pan_dict


def convert_dates(pan_dict, dates):
    for k in pan_dict.keys():
        df = pan_dict[k]
        for date_field in dates[k]:
            df[date_field] = pd.to_datetime(df[date_field], format="%d/%m/%Y").dt.date
        pan_dict[k] = df
    return pan_dict


def export_file(output, pan_dict):
    """Writes output file as xlsx"""
    output_file = Path(output, f"pan_London_Annex_A.xlsx")
    with pd.ExcelWriter(output_file) as writer:
        for k in pan_dict.keys():
            df = pan_dict[k]
            df.to_excel(writer, sheet_name=k, index=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['liiatools',
 'liiatools.csdatatools',
 'liiatools.csdatatools.datasets',
 'liiatools.csdatatools.datasets.cincensus',
 'liiatools.csdatatools.util',
 'liiatools.datasets',
 'liiatools.datasets.annex_a',
 'liiatools.datasets.annex_a.lds_annexa_clean',
 'liiatools.datasets.annex_a.lds_annexa_la_agg',
 'liiatools.datasets.annex_a.lds_annexa_pan_agg',
 'liiatools.datasets.cin_census',
 'liiatools.datasets.cin_census.lds_cin_clean',
 'liiatools.datasets.cin_census.lds_cin_la_agg',
 'liiatools.datasets.cin_census.lds_cin_pan_agg',
 'liiatools.datasets.s903',
 'liiatools.datasets.s903.lds_ssda903_clean',
 'liiatools.datasets.s903.lds_ssda903_la_agg',
 'liiatools.datasets.s903.lds_ssda903_pan_agg',
 'liiatools.datasets.s903.lds_ssda903_sufficiency',
 'liiatools.datasets.school_census',
 'liiatools.datasets.school_census.lds_school_clean',
 'liiatools.datasets.school_census.lds_school_la_agg',
 'liiatools.datasets.shared_functions',
 'liiatools.spec',
 'liiatools.spec.annex_a',
 'liiatools.spec.cin_census',
 'liiatools.spec.common',
 'liiatools.spec.s903',
 'liiatools.spec.school_census']

package_data = \
{'': ['*'],
 'liiatools.spec.annex_a': ['samples/*'],
 'liiatools.spec.cin_census': ['samples/*'],
 'liiatools.spec.s903': ['samples/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'XlsxWriter>=3.0.3,<4.0.0',
 'cchardet==2.1.7',
 'click-log>=0.4.0,<0.5.0',
 'click>=8.1.2,<9.0.0',
 'dacite>=1.6.0,<2.0.0',
 'lxml>=4.9.1,<5.0.0',
 'more-itertools>=8.12.0,<9.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'regex>=2022.4.24,<2023.0.0',
 'sfdata-stream-parser==0.4.1',
 'tablib[cli,xlsx]>=3.2.0,<4.0.0',
 'xlrd>=2.0.1,<3.0.0',
 'xmlschema>=1.10.0,<2.0.0']

setup_kwargs = {
    'name': 'liiatools',
    'version': '0.1.4.7',
    'description': "Children's Services Data Tools - Utilities for cleaning and normalising CS data by Social Finance",
    'long_description': 'None',
    'author': 'Michael Hanks',
    'author_email': 'michael.hanks@socialfinance.org.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
import pandas as pd
import logging

from liiatools.datasets.shared_functions import converters, common

log = logging.getLogger(__name__)


def convert_to_dataframe(data):
    data = data.export("df")
    return data


def get_year(input, data, la_log_dir):
    filename = Path(input).stem
    try:
        year = common.check_year(filename)
        data["YEAR"] = year
        return data
    except AttributeError:
        common.save_year_error(input, la_log_dir)
        raise Exception(
            f"{filename} was not processed as the filename did not contain a year, "
            f"this error has been sent to the LA to be fixed"
        )


def convert_to_datetime(data):
    data[["PersonBirthDate", "PersonDeathDate", "ExpectedPersonBirthDate"]] = data[
        ["PersonBirthDate", "PersonDeathDate", "ExpectedPersonBirthDate"]
    ].apply(pd.to_datetime)
    return data


def _get_person_school_year(datevalue):
    if datevalue.month >= 9:
        school_year = datevalue.year
    elif datevalue.month <= 8:
        school_year = datevalue.year - 1
    else:
        school_year = None
    return school_year


def add_school_year(data):
    data["PersonSchoolYear"] = data["PersonBirthDate"].apply(
        lambda row: _get_person_school_year(row)
    )
    return data


def add_la_name(data, la_name):
    data["LA"] = la_name
    return data


def la_prefix(data, la_code):
    data["LAchildID"] = data["LAchildID"] + "_" + la_code
    return data


def degrade_dob(data):
    if data["PersonBirthDate"] is not None:
        data["PersonBirthDate"] = data["PersonBirthDate"].apply(
            lambda row: converters.to_month_only_dob(row)
        )
        return data


def degrade_expected_dob(data):
    if data["ExpectedPersonBirthDate"] is not None:
        data["ExpectedPersonBirthDate"] = data["ExpectedPersonBirthDate"].apply(
            lambda row: converters.to_month_only_dob(row)
        )
        return data


def degrade_death_date(data):
    if data["PersonDeathDate"] is not None:
        data["PersonDeathDate"] = data["PersonDeathDate"].apply(
            lambda row: converters.to_month_only_dob(row)
        )
        return data


def add_fields(input, data, la_name, la_log_dir, la_code):
    """Adds YEAR, LA, PERSONSCHOOLYEAR to exported dataframe
    Appends LA_code from config to LAChildID"""
    data = convert_to_dataframe(data)
    data = get_year(input, data, la_log_dir)
    data = convert_to_datetime(data)
    data = add_school_year(data)
    data = add_la_name(data, la_name)
    data = la_prefix(data, la_code)
    data = degrade_dob(data)
    data = degrade_expected_dob(data)
    data = degrade_death_date(data)
    return data


def export_file(input, output, data):
    filename = Path(input).stem
    outfile = filename + "_clean.csv"
    output_path = Path(output, outfile)
    data.to_csv(output_path, index=False)



from pathlib import Path
import pandas as pd
import logging

log = logging.getLogger(__name__)


def read_file(file):
    """
    Reads the csv file as a pandas DataFrame
    """
    filepath = Path(file)
    s903_df = pd.read_csv(filepath, index_col=None)
    return s903_df


def match_load_file(s903_df, column_names):
    """
    Matches the columns in the DataFrame against one of the 10 SSDA903 file types
    """
    for table_name, expected_columns in column_names.items():
        if set(s903_df.columns) == set(expected_columns):
            return table_name


def data_min(s903_df, minimise, table_name):
    """
    Performs additional data minimisation for pan-London sufficiency analysis in line with schema
    """
    if table_name in minimise.keys():
        s903_df = s903_df.drop(columns=minimise[table_name], axis=1)
    return s903_df


def export_suff_file(output, table_name, s903_df):
    """
    Writes file to output directory
    """
    output_path = Path(output, f"pan_London_SSDA903_{table_name}.csv")
    s903_df.to_csv(output_path, index=False)
